import logging
from os import environ
from typing import Optional
from warnings import warn

import zenoh

_maybe_session: Optional[zenoh.Session] = None

logger = logging.getLogger(__name__)


def set_auto_session(session: Optional[zenoh.Session] = None) -> None:
    """Set the global shared session instance.

    If called with a session, it replaces the current global session.
    If called with None, the global session is unset (but not close!).

    Args:
        session: zenoh session to set as default
    """
    global _maybe_session
    _maybe_session = session

def auto_session(session: Optional[zenoh.Session] = None) -> zenoh.Session:
    """Returns the passed session.
    If None, returns a singleton session shared with every other None call of
    this function."""
    global _maybe_session
    if _maybe_session is not None:
        return _maybe_session

    if "ZENOH_SESSION_CONFIG_URI" in environ:
        ZENOH_CONFIG = zenoh.Config.from_file(environ["ZENOH_SESSION_CONFIG_URI"])
    else:
        warn(f"'ZENOH_SESSION_CONFIG_URI' environment variable is not set. Using default session provided by zenoh")
        ZENOH_CONFIG = zenoh.Config()
    logger.info("Global zenoh session: Starting")
    ses = zenoh.open(ZENOH_CONFIG)
    set_auto_session(ses)
    logger.info("Global zenoh session: Running")
    return ses



import asyncio
import copy
import logging
from typing import Any, AsyncGenerator, Generator


import pytest
pytest.importorskip("rclpy")
import rclpy
from rclpy.publisher import Publisher
from rclpy.qos import QoSProfile
from std_msgs.msg import String

import asyncio_for_robotics.ros2 as aros
from asyncio_for_robotics import soft_wait_for
from asyncio_for_robotics.core._logger import setup_logger
from asyncio_for_robotics.core.utils import soft_timeout
from asyncio_for_robotics.ros2.session import ThreadedSession, set_auto_session

setup_logger(debug_path="tests")
logger = logging.getLogger("asyncio_for_robotics.test")


@pytest.fixture(scope="module", autouse=True)
def session() -> Generator[aros.BaseSession, Any, Any]:
    logger.info("Starting rclpy and session")
    rclpy.init()
    set_auto_session(ThreadedSession())
    ses = aros.auto_session()
    yield ses
    logger.info("closing rclpy and session")
    ses.close()
    rclpy.shutdown()


topic = aros.TopicInfo(
    "test/something",
    String,
    QoSProfile(
        depth=10000,
    ),
)


@pytest.fixture(scope="module")
def pub(session: aros.BaseSession) -> Generator[Publisher, Any, Any]:
    with session.lock() as node:
        p: Publisher = node.create_publisher(**topic.as_kwarg())
    yield p
    with session.lock() as node:
        node.destroy_publisher(p)


@pytest.fixture
async def sub(session) -> AsyncGenerator[aros.Sub[String], Any]:
    s: aros.Sub = aros.Sub(**topic.as_kwarg())
    yield s
    s.close()


async def test_wait_for_value(pub: Publisher, sub: aros.Sub[String]):
    logger.info("entered test")
    payload = "hello"
    logger.info("publishing")
    pub.publish(String(data=payload))
    logger.info("awaiting data")
    sample = await soft_wait_for(sub.wait_for_value(), 1)
    assert not isinstance(sample, TimeoutError), f"Did not receive response in time"
    assert sample.data == payload


async def test_wait_new(pub: Publisher, sub: aros.Sub[String]):
    payload = "hello"
    pub.publish(String(data=payload))
    sample = await sub.wait_for_value()
    assert not isinstance(sample, TimeoutError), f"Should get a message"

    wait_task = sub.wait_for_new()
    new_sample = await soft_wait_for(wait_task, 0.1)
    assert isinstance(new_sample, TimeoutError), f"Should not get a message"

    wait_task = sub.wait_for_new()
    pub.publish(String(data=payload))
    new_sample = await soft_wait_for(wait_task, 0.1)
    assert not isinstance(new_sample, TimeoutError), f"Should get the message"
    assert new_sample.data == payload


async def test_wait_next(pub: Publisher, sub: aros.Sub[String]):
    first_payload = "hello"
    pub.publish(String(data=first_payload))
    sample = await sub.wait_for_value()
    assert not isinstance(sample, TimeoutError), f"Should get a message"

    wait_task = sub.wait_for_next()
    new_sample = await soft_wait_for(wait_task, 0.1)
    assert isinstance(new_sample, TimeoutError), f"Should not get a message"

    wait_task = sub.wait_for_next()
    pub.publish(String(data=first_payload))
    for other_payload in range(10):
        await asyncio.sleep(0.001)
        pub.publish(String(data=str(other_payload)))

    new_sample = await soft_wait_for(wait_task, 0.1)
    assert not isinstance(new_sample, TimeoutError), f"Should get the message"
    assert new_sample.data == first_payload


async def test_listen_one_by_one(pub: Publisher, sub: aros.Sub[String]):
    last_payload = "hello"
    pub.publish(String(data=last_payload))
    sample_count = 0
    put_count = 1
    max_iter = 20
    async for sample in sub.listen():
        sample_count += 1
        assert sample.data == last_payload
        if sample_count >= max_iter:
            break
        last_payload = f"hello{sample_count}"
        pub.publish(String(data=last_payload))
        put_count += 1

    assert put_count == sample_count == max_iter


async def test_listen_too_fast(pub: Publisher, sub: aros.Sub[String]):
    last_payload = "hello"
    pub.publish(String(data=last_payload))
    pub.publish(String(data=last_payload))
    sample_count = 0
    put_count = 2
    max_iter = 20
    await asyncio.sleep(0.01)
    async for sample in sub.listen():
        sample_count += 1
        assert sample.data == last_payload
        if sample_count >= max_iter:
            break
        last_payload = f"hello{sample_count}"
        pub.publish(String(data=last_payload))
        put_count += 1
        last_payload = f"hello{sample_count}"
        pub.publish(String(data=last_payload))
        put_count += 1
        await asyncio.sleep(0.01)

    assert put_count / 2 == sample_count == max_iter


async def test_reliable_one_by_one(pub: Publisher, sub: aros.Sub[String]):
    last_payload = "hello"
    pub.publish(String(data=last_payload))
    sample_count = 0
    put_count = 1
    max_iter = 20
    async for sample in sub.listen_reliable():
        sample_count += 1
        assert sample.data == last_payload
        if sample_count >= max_iter:
            break
        last_payload = f"hello{sample_count}"
        pub.publish(String(data=last_payload))
        put_count += 1

    assert put_count == sample_count == max_iter


async def test_reliable_too_fast(pub: Publisher, sub: aros.Sub[String]):
    data = list(range(200))
    put_queue = [str(v) for v in data]
    received_buf = []
    listener = sub.listen_reliable(fresh=True, queue_size=len(data) * 2)
    await asyncio.sleep(0.1)
    pub.publish(String(data=put_queue.pop()))
    pub.publish(String(data=put_queue.pop()))
    async with soft_timeout(2):
        async for sample in listener:
            payload = int(sample.data)
            received_buf.append(payload)
            if len(received_buf) >= len(data):
                break
            if put_queue != []:
                pub.publish(String(data=put_queue.pop()))
            if put_queue != []:
                pub.publish(String(data=put_queue.pop()))

    received_buf.reverse()
    assert data == received_buf


async def test_freshness(pub: Publisher, sub: aros.Sub[String]):
    payload = "hello"
    new = sub.wait_for_new()
    pub.publish(String(data=payload))
    await new
    sample = await soft_wait_for(anext(sub.listen(fresh=False)), 0.1)
    assert not isinstance(sample, TimeoutError), f"Should get the message"
    assert sample.data == payload

    new = sub.wait_for_new()
    pub.publish(String(data=payload))
    await new
    sample = await soft_wait_for(anext(sub.listen_reliable(fresh=False)), 0.1)
    assert not isinstance(sample, TimeoutError), f"Should get the message"
    assert sample.data == payload
    await sub.wait_for_value()

    new = sub.wait_for_new()
    pub.publish(String(data=payload))
    await new
    sample = await soft_wait_for(anext(sub.listen(fresh=True)), 0.1)
    assert isinstance(sample, TimeoutError), f"Should NOT get the message. got {sample}"

    new = sub.wait_for_new()
    pub.publish(String(data=payload))
    await new
    sample = await soft_wait_for(anext(sub.listen_reliable(fresh=True)), 0.1)
    assert isinstance(sample, TimeoutError), f"Should NOT get the message"



"""Implements a simple ros2 talker, publishing in a non-blocking asyncio loop"""

import asyncio
from contextlib import suppress

import rclpy
from std_msgs.msg import String

from asyncio_for_robotics.core.utils import Rate
from asyncio_for_robotics.ros2 import TopicInfo, auto_session

TOPIC = TopicInfo(msg_type=String, topic="example/talker")


async def main():
    # create the publisher safely
    with auto_session().lock() as node:
        pub = node.create_publisher(TOPIC.msg_type, TOPIC.topic, TOPIC.qos)

    count = 0
    async for t in Rate(frequency=2).listen_reliable():  # stable timer
        data = f"[Hello world! timestamp: {count/10:.1f}s]"
        count += 1
        print(f"Sending: {data}")
        pub.publish(String(data=data))  # sends data (lock is not necessary)
        await asyncio.sleep(0.5)  # non-blocking sleep


if __name__ == "__main__":
    rclpy.init()
    try:
        # suppress, just so we don't flood the terminal on exit
        with suppress(KeyboardInterrupt, asyncio.CancelledError):
            asyncio.run(main())  # starts asyncio executor
    finally:
        # cleanup. `finally` statment always executes
        auto_session().close()
        rclpy.shutdown()



import asyncio
from contextlib import suppress

import rclpy
from example_interfaces.srv import AddTwoInts

import asyncio_for_robotics.ros2 as afor

TOPIC = afor.TopicInfo("add_two_ints", AddTwoInts)


async def fibo_server():
    server = afor.Server(*TOPIC.as_arg())
    print(f"Awaiting request")
    async for responder in server.listen_reliable():
        # responder object contains the request
        a = responder.request.a
        b = responder.request.b
        # responder object also contains the response field that we should set
        responder.response.sum = a + b
        # we send the response
        responder.send()
        print(f"Replied: {a} + {b} = {a+b}")


if __name__ == "__main__":
    rclpy.init()
    try:
        # suppress, just so we don't flood the terminal on exit
        with suppress(KeyboardInterrupt, asyncio.CancelledError):
            asyncio.run(fibo_server())  # starts asyncio executor
    finally:
        # cleanup. `finally` statment always executes
        afor.auto_session().close()
        rclpy.shutdown()



import logging

import pytest

pytest.importorskip("zenoh")

import asyncio
from contextlib import suppress
from typing import Any, AsyncGenerator, Generator

import zenoh

from asyncio_for_robotics.zenoh import Sub, auto_session, soft_timeout, soft_wait_for

logger = logging.getLogger("asyncio_for_robotics.test")


@pytest.fixture(scope="module", autouse=True)
def session() -> Generator[zenoh.Session, Any, Any]:
    ses = auto_session()
    yield ses
    if not auto_session().is_closed():
        with suppress(zenoh.ZError):
            ses.close()


@pytest.fixture
def pub(session) -> Generator[zenoh.Publisher, Any, Any]:
    pub_topic = "test/something"
    logger.debug("Creating PUB-%s", pub_topic)
    p: zenoh.Publisher = auto_session().declare_publisher(pub_topic)
    yield p
    if not auto_session().is_closed():
        logger.debug("closing PUB-%s", pub_topic)
        p.undeclare()


@pytest.fixture
async def sub(session) -> AsyncGenerator[Sub, Any]:
    s: Sub = Sub("test/**")
    yield s
    s.close()


async def test_wait_for_value(pub: zenoh.Publisher, sub: Sub):
    payload = b"hello"
    pub.put(payload)
    sample = await soft_wait_for(sub.wait_for_value(), 1)
    assert not isinstance(sample, TimeoutError), f"Did not receive response in time"
    assert sample.payload.to_bytes() == payload
    assert str(sample.key_expr) == str(pub.key_expr)


async def test_wait_new(pub: zenoh.Publisher, sub: Sub):
    payload = b"hello"
    pub.put(payload)
    sample = await sub.wait_for_value()
    assert not isinstance(sample, TimeoutError), f"Should get a message"

    wait_task = sub.wait_for_new()
    new_sample = await soft_wait_for(wait_task, 0.1)
    assert isinstance(new_sample, TimeoutError), f"Should not get a message"

    wait_task = sub.wait_for_new()
    pub.put(payload)
    new_sample = await soft_wait_for(wait_task, 0.1)
    assert not isinstance(new_sample, TimeoutError), f"Should get the message"
    assert new_sample.payload.to_bytes() == payload


async def test_wait_next(pub: zenoh.Publisher, sub: Sub):
    first_payload = b"hello"
    pub.put(first_payload)
    sample = await sub.wait_for_value()
    assert not isinstance(sample, TimeoutError), f"Should get a message"

    wait_task = sub.wait_for_next()
    new_sample = await soft_wait_for(wait_task, 0.1)
    assert isinstance(new_sample, TimeoutError), f"Should not get a message"

    wait_task = sub.wait_for_next()
    pub.put(first_payload)
    for other_payload in range(10):
        await asyncio.sleep(0.001)
        pub.put(str(other_payload))

    new_sample = await soft_wait_for(wait_task, 0.1)
    assert not isinstance(new_sample, TimeoutError), f"Should get the message"
    assert new_sample.payload.to_bytes() == first_payload


async def test_listen_one_by_one(pub: zenoh.Publisher, sub: Sub):
    last_payload = "hello"
    pub.put(last_payload)
    sample_count = 0
    put_count = 1
    max_iter = 20
    async for sample in sub.listen():
        sample_count += 1
        assert sample.payload.to_string() == last_payload
        if sample_count >= max_iter:
            break
        last_payload = f"hello{sample_count}"
        pub.put(last_payload)
        put_count += 1

    assert put_count == sample_count == max_iter


async def test_listen_too_fast(pub: zenoh.Publisher, sub: Sub):
    last_payload = "hello"
    pub.put(last_payload)
    pub.put(last_payload)
    sample_count = 0
    put_count = 2
    max_iter = 20
    await asyncio.sleep(0.01)
    async for sample in sub.listen():
        sample_count += 1
        assert sample.payload.to_string() == last_payload
        if sample_count >= max_iter:
            break
        last_payload = f"hello{sample_count}"
        pub.put(last_payload)
        put_count += 1
        last_payload = f"hello{sample_count}"
        pub.put(last_payload)
        put_count += 1
        await asyncio.sleep(0.01)

    assert put_count / 2 == sample_count == max_iter


async def test_reliable_one_by_one(pub: zenoh.Publisher, sub: Sub):
    last_payload = "hello"
    pub.put(last_payload)
    sample_count = 0
    put_count = 1
    max_iter = 20
    async for sample in sub.listen_reliable():
        sample_count += 1
        assert sample.payload.to_string() == last_payload
        if sample_count >= max_iter:
            break
        last_payload = f"hello{sample_count}"
        pub.put(last_payload)
        put_count += 1

    assert put_count == sample_count == max_iter


async def test_reliable_too_fast(pub: zenoh.Publisher, sub: Sub):
    data = list(range(200))
    put_queue = [str(v) for v in data]
    received_buf = []
    listener = sub.listen_reliable(fresh=True, queue_size=len(data))
    await asyncio.sleep(0.1)
    pub.put(put_queue.pop())
    pub.put(put_queue.pop())
    async with soft_timeout(2):
        async for sample in listener:
            payload = int(sample.payload.to_string())
            received_buf.append(payload)
            if len(received_buf) >= len(data):
                break
            if put_queue != []:
                pub.put(put_queue.pop())
            if put_queue != []:
                pub.put(put_queue.pop())

    received_buf.reverse()
    assert data == received_buf


async def test_freshness(pub: zenoh.Publisher, sub: Sub):
    payload = "hello"

    new = sub.wait_for_new()
    pub.put(payload)
    await new
    sample = await soft_wait_for(anext(sub.listen(fresh=False)), 0.1)
    assert not isinstance(sample, TimeoutError), f"Should get the message"
    assert sample.payload.to_string() == payload

    new = sub.wait_for_new()
    pub.put(payload)
    await new
    sample = await soft_wait_for(anext(sub.listen_reliable(fresh=False)), 0.1)
    assert not isinstance(sample, TimeoutError), f"Should get the message"
    assert sample.payload.to_string() == payload
    await sub.wait_for_value()

    new = sub.wait_for_new()
    pub.put(payload)
    await new
    sample = await soft_wait_for(anext(sub.listen(fresh=True)), 0.1)
    assert isinstance(sample, TimeoutError), f"Should NOT get the message"

    new = sub.wait_for_new()
    pub.put(payload)
    await new
    sample = await soft_wait_for(anext(sub.listen_reliable(fresh=True)), 0.1)
    assert isinstance(sample, TimeoutError), f"Should NOT get the message"



import json
import copy
import logging.config
import os
from datetime import datetime
from typing import Optional

import colorama
from colorama import Fore, Style, init


class JsonLineFormatter(logging.Formatter):
    """Outputs log records as JSON lines."""

    def format(self, record):
        log_record = {
            "time": datetime.fromtimestamp(record.created).isoformat(),
            "level": record.levelname,
            "logger": record.name,
            "message": record.getMessage(),
            "module": record.module,
            "funcName": record.funcName,
            "line": record.lineno,
        }
        if record.exc_info:
            log_record["exc_info"] = self.formatException(record.exc_info)
        return json.dumps(log_record)


LEVEL_COLORS = {
    "DEBUG": Fore.BLUE,
    "INFO": Fore.CYAN,
    "WARNING": Fore.YELLOW,
    "ERROR": Fore.RED,
    "CRITICAL": Fore.BLACK + Style.BRIGHT + colorama.Back.RED,
}


class OnlyLevelFilter(logging.Filter):
    """Only allow a single level through this handler."""
    def __init__(self, level):
        self.level = level
    def filter(self, record):
        return record.levelno == self.level

class ColoredFormatter(logging.Formatter):
    """Adds colors to levelname in logs."""

    def format(self, record: logging.LogRecord):
        record = copy.deepcopy(record)
        levelname = record.levelname
        if levelname in LEVEL_COLORS:
            record.levelname = (
                f"{LEVEL_COLORS[levelname]}{record.levelname[0]}{Style.RESET_ALL}"
            )
            record.msg = f"{LEVEL_COLORS[levelname]}{record.msg}{Style.RESET_ALL}"
        return super().format(record)


def setup_logger(debug_path: Optional[str] = None):
    handlers = ["stdout", "stderr"]
    if debug_path:
        handlers.append("json")
        handlers.append("userlog")
    cfg = {
        "version": 1,
        "disable_existing_loggers": False,
        "formatters": {
            "user": {
                "()": ColoredFormatter,
                "format": "%(levelname)s| %(message)s",
                "datefmt": "%Y-%m-%dT%H:%M:%S",
            },
            "json": {
                "()": JsonLineFormatter,  # custom formatter
            },
        },
        "filters": {"only_info": {"()": OnlyLevelFilter, "level": logging.INFO}},
        "handlers": {
            "stdout": {
                "class": "logging.StreamHandler",
                "level": "INFO",
                "formatter": "user",
                "filters": ["only_info"],
                "stream": "ext://sys.stdout",
            },
            "stderr": {
                "class": "logging.StreamHandler",
                "level": "WARNING",
                "formatter": "user",
                "stream": "ext://sys.stderr",
            },
            "json": {
                "class": "logging.FileHandler",
                "level": "DEBUG",
                "formatter": "json",
                "filename": (
                    os.path.expanduser(debug_path) + "/debug.log.jsonl"
                    if debug_path is not None
                    else "log.jsonl"
                ),  # path relative to working dir
                "mode": "w",
            },
            "userlog": {
                "class": "logging.FileHandler",
                "level": "DEBUG",
                "formatter": "user",
                "filename": (
                    os.path.expanduser(debug_path) + "/debug.log"
                    if debug_path is not None
                    else "log"
                ),  # path relative to working dir
                "mode": "w",
            },
        },
        "loggers": {
            "asyncio_for_robotics": {
                "level": "DEBUG",
                "handlers": handlers,
                "propagate": False,
            },
        },
    }
    logging.config.dictConfig(cfg)




import logging
from typing import Any, AsyncGenerator, Generator

import pytest
pytest.importorskip("rclpy")
pytest.importorskip("yaml")
import rclpy
from rclpy.publisher import Publisher
from rclpy.qos import QoSProfile
from std_msgs.msg import String
from test_ros2_pubsub_thr import (
    test_freshness,
    test_listen_one_by_one,
    test_listen_too_fast,
    test_reliable_one_by_one,
    test_reliable_too_fast,
    test_wait_for_value,
    test_wait_new,
    test_wait_next,
)

import asyncio_for_robotics.ros2 as aros
from asyncio_for_robotics.core._logger import setup_logger
from asyncio_for_robotics.ros2.session import SynchronousSession, set_auto_session

logger = logging.getLogger("asyncio_for_robotics.test")


@pytest.fixture(scope="module", autouse=True)
def rclpy_init():
    logger.info("Starting rclpy")
    rclpy.init()
    yield
    logger.info("closing rclpy ")
    rclpy.shutdown()


@pytest.fixture(scope="function", autouse=True)
async def session(rclpy_init) -> AsyncGenerator[aros.BaseSession, Any]:
    logger.info("Starting session")
    set_auto_session(SynchronousSession())
    ses = aros.auto_session()
    yield ses
    logger.info("closing session")
    ses.close()


TOPIC = aros.TopicInfo(
    "test/something",
    String,
    QoSProfile(
        depth=100,
    ),
)


@pytest.fixture
def pub(session: aros.BaseSession) -> Generator[Publisher, Any, Any]:
    with session.lock() as node:
        p: Publisher = node.create_publisher(**TOPIC.as_kwarg())
    yield p
    with session.lock() as node:
        node.destroy_publisher(p)


@pytest.fixture
async def sub(session) -> AsyncGenerator[aros.Sub[String], Any]:
    s: aros.Sub = aros.Sub(**TOPIC.as_kwarg())
    yield s
    s.close()



import asyncio
import copy
import logging
from typing import Any, AsyncGenerator, Generator

import pytest

pytest.importorskip("rclpy")
pytest.importorskip("yaml")
import rclpy
from rclpy.publisher import Publisher
from rclpy.qos import QoSProfile
from std_msgs.msg import String
from std_srvs.srv import SetBool

import asyncio_for_robotics.ros2 as aros
from asyncio_for_robotics import soft_wait_for
from asyncio_for_robotics.core._logger import setup_logger
from asyncio_for_robotics.ros2.service import Client, Server
from asyncio_for_robotics.ros2.session import ThreadedSession, set_auto_session

setup_logger(debug_path="tests")
logger = logging.getLogger("asyncio_for_robotics.test")


@pytest.fixture(scope="module", autouse=True)
def session() -> Generator[aros.BaseSession, Any, Any]:
    logger.info("Starting rclpy and session")
    rclpy.init()
    set_auto_session(ThreadedSession())
    ses = aros.auto_session()
    yield ses
    logger.info("closing rclpy and session")
    ses.close()
    rclpy.shutdown()


topic = aros.TopicInfo(
    "test/something",
    String,
    QoSProfile(
        depth=100,
    ),
)


@pytest.fixture
async def server(
    session: aros.BaseSession,
) -> AsyncGenerator[Server[SetBool.Request, SetBool.Response], Any]:
    server = Server(SetBool, "test/srv")
    yield server
    server.close()


@pytest.fixture
async def client(
    session: aros.BaseSession,
) -> AsyncGenerator[Client[SetBool.Request, SetBool.Response], Any]:
    client = Client(SetBool, "test/srv")
    yield client
    client.close()


async def test_client_receives_response(
    server: Server[SetBool.Request, SetBool.Response],
    client: Client[SetBool.Request, SetBool.Response],
):
    response_async = client.call(SetBool.Request(data=True))

    responder = await soft_wait_for(server.wait_for_value(), 1)
    assert not isinstance(responder, TimeoutError), f"Server did not receive request"
    assert responder.request.data == True
    responder.response.success = True
    responder.response.message = "hello"
    responder.send()

    response = await soft_wait_for(response_async, 1)
    assert not isinstance(response, TimeoutError), f"Client did not receive reply"
    assert response.message == "hello"
    assert response.success == True
    return



import asyncio
import logging
from abc import ABC, abstractmethod
from asyncio.queues import Queue
from collections import deque
from types import CoroutineType
from typing import (
    Any,
    AsyncGenerator,
    Awaitable,
    Callable,
    Coroutine,
    Deque,
    Final,
    Generic,
    List,
    Optional,
    Set,
    TypeVar,
)

logger = logging.getLogger(__name__)

_MsgType = TypeVar("_MsgType")


class BaseSub(Generic[_MsgType], ABC):
    def __init__(
        self,
    ) -> None:
        """Abstract (non-specific) asyncio_for_robotics subscriber.

        This defines the different methods to asynchronously get new messages.

        Use ``self.intput_data(your_data)`` to input new messages on this subscriber.

        This object is pure python and not specialize for any transport
        protocol. See asyncio_for_robotics.zenoh.sub for an easy
        implementation.

        To implements your own sheduling methods (new type of queue, buffer,
        genrator ...), please either: 

            - inherit from this class, then overide self._input_data_asyncio
            - put a callback inside self.asap_callback
        """
        #: Blocking callbacks called on message arrival
        self.asap_callback: List[Callable[[_MsgType], Any]] = []
        #: Event triggering when the first message is received
        self.alive: asyncio.Event = asyncio.Event()
        #: Number of messages received since start
        self.msg_count: int = 0
        #: Condition that fires on new data
        self.new_value_cond = asyncio.Condition()
        #: queues associated with the generators
        self._dyncamic_queues: Set[Queue[_MsgType]] = set()

        self._event_loop: asyncio.AbstractEventLoop = asyncio.get_event_loop()
        #: Value is available event
        self._value_flag: asyncio.Event = asyncio.Event()
        #: Lastest message value
        self._value: Optional[_MsgType] = None
        logger.debug("created sub %s", self.name)

    @property
    def name(self) -> str:
        """The friendly name of you subscriber"""
        return "no_name"

    def input_data(self, data: _MsgType) -> bool:
        """Processes incomming data.
        Call this in your subscriber callback.

        .. Note:
            This is threadsafe, thus can run safely on any thread.

        Args:
            data: Data to input on this sub

        Returns:
            False if the even loop has been closed or there is another critical
                problem making this sub unable to work.
        """
        if self._event_loop.is_closed():
            logger.info("Event loop closed, for sub: %s", self.name)
            return False
        self._event_loop.call_soon_threadsafe(self._input_data_asyncio, data)
        return True


    async def wait_for_value(self) -> _MsgType:
        """Latest message.

        Returns:
            The latest message (if none, awaits the first message)
        """
        await self._value_flag.wait()
        assert self._value is not None
        return self._value

    def wait_for_new(self) -> Coroutine[Any, Any, _MsgType]:
        """Awaits a new value.

        .. Note:
            "Listening" starts the moment this function is called, not when
            the coroutine is awaited

        See wait_for_next to be sure the data is exactly the next received.

        Returns:
            Coroutine holding a message more recent than the one at time of the call
        """
        last_count = self.msg_count

        async def func() -> _MsgType:
            async with self.new_value_cond:
                await self.new_value_cond.wait_for(lambda: self.msg_count > last_count)
            assert self._value is not None
            return self._value

        return func()

    def wait_for_next(self) -> Awaitable[_MsgType]:
        """Awaits exactly the next value.

        .. Note:
            "Listening" starts the moment this function is called, not when
            the coroutine is awaited

        Returns:
            Coroutine holding the first message received after the call.
        """
        q: asyncio.Queue[_MsgType] = asyncio.LifoQueue(maxsize=2)
        self._dyncamic_queues.add(q)

        async def func() -> _MsgType:
            try:
                val_top = await q.get()
                if q.empty():
                    val_deep = val_top
                else:
                    val_deep = q.get_nowait()
                return val_deep
            finally:
                self._dyncamic_queues.discard(q)

        return func()

    def listen(self, fresh=False) -> AsyncGenerator[_MsgType, None]:
        """Itterates over the newest message.

        Messages might be skipped, this is not a queue.

        .. Note:
            "Listening" starts the moment this function is called, not when
            the generator is awaited

        Args:
            fresh: If false, first yield can be the latest value

        Returns:
            Async generator itterating over the newest message.
        """
        return self.listen_reliable(fresh, 1, False)

    def listen_reliable(
        self, fresh=False, queue_size: int = 10, lifo=False
    ) -> AsyncGenerator[_MsgType, None]:
        """Itterates over every incomming messages. (does not miss messages)

        .. Note:
            "Listening" starts the moment this function is called, not when
            the generator is awaited.


        Args:
            fresh: If false, first yield can be the latest value
            queue_size: size of the queue of values
            lifo: If True, uses a last in first out queue instead of default fifo.

        Returns:
            Async generator itterating over every incomming message.
        """
        if not lifo:
            q: asyncio.Queue[_MsgType] = asyncio.Queue(maxsize=queue_size)
        else:
            q: asyncio.Queue[_MsgType] = asyncio.LifoQueue(maxsize=queue_size)
        self._dyncamic_queues.add(q)
        logger.debug("Reliable listener primed %s", self.name)
        if self._value_flag.is_set() and not fresh:
            assert self._value is not None, "impossible if flag set"
            q.put_nowait(self._value)
        return self._unprimed_listen_reliable(q)

    async def _unprimed_listen_reliable(
        self, queue: asyncio.Queue
    ) -> AsyncGenerator[_MsgType, None]:
        logger.debug("Reliable listener first iter %s", self.name)
        try:
            while True:
                # logger.debug("Reliable listener waiting data %s", self.name)
                msg = await queue.get()
                # logger.debug("Reliable listener got data %s", self.name)
                yield msg
                # logger.debug("Reliable listener yielded data %s", self.name)
        finally:
            self._dyncamic_queues.discard(queue)
            logger.debug("Reliable listener closed %s", self.name)

    def _input_data_asyncio(self, msg: _MsgType):
        """Processes incomming data.

        Is only safe to run on the same thread as asyncio.

        Args:
            msg: message
        """
        # logger.debug("Input message %s", self.name)
        for f in self.asap_callback:
            f(msg)
        for q in self._dyncamic_queues:
            if q.full():
                if q.qsize() > 2:
                    logger.warning("Queue full (%s) on %s", q.qsize, self.name)
                q.get_nowait()
            q.put_nowait(msg)
        self._value = msg
        self._value_flag.set()
        self.msg_count += 1
        asyncio.create_task(self._wakeup_new())
        if not self.alive.is_set():
            logger.debug("%s is receiving data", self.name)
            self.alive.set()

    async def _wakeup_new(self):
        """fires new_value_cond"""
        async with self.new_value_cond:
            self.new_value_cond.notify_all()
